
#ifndef XiaomiDevice_h
#define XiaomiDevice_h

#include "Arduino.h"
#include <ArduinoJson.h>
#include <ArduinoLog.h>
#include <AES.h>
#include <ctype.h>
#include <ESP8266WiFi.h>
#include <WiFiUDP.h>

#define VERSION "1.0.0"

// Multicast ip address + port definition
const IPAddress MULTICAST_ADDRESS(224, 0, 0, 50);
const unsigned int MULTICAST_PORT = 9898;
const unsigned int GATEWAY_DISCOVERY_PORT = 4321;

// Maximum supported gateways
#define MAX_GATEWAYS 4

// Maximum allowed devices per gateway
#define MAX_DEVICES 32

// Delays to receive the packet
#define DISCOVERY_DELAY 5000
#define RESPONSE_DELAY 5000

// Define max sizes
#define MAX_PACKET_SIZE 512
#define MAX_JSON_DICT_SIZE  (JSON_OBJECT_SIZE(16) + MAX_PACKET_SIZE)
#define MAX_JSON_ARRAY_SIZE (JSON_ARRAY_SIZE(64) + MAX_PACKET_SIZE)

// Structure for definition of gateways
struct XIAOMI_GATEWAYS {
  String ip;
  String password;
};

// RGB for gateway
struct RGB {
   uint8_t r;
   uint8_t g;
   uint8_t b;
};

// Xiaomi commands
#define MAX_XIAOMI_COMMAND_PARAMS 8
struct XIAOMI_COMMAND {
  String command;
  uint8_t numParams = 0;
  String params[MAX_XIAOMI_COMMAND_PARAMS];
};

// Callback
typedef void (*XiaomiCallback) (String model, String sid, String key, String value);

// Helper functions
boolean isValidNumber(String str);

// Class to send UDP packets
class XiaomiUDP {
  public:
    XiaomiUDP(WiFiUDP udp, IPAddress ip, uint16_t port): _udp(udp), _ip(ip), _port(port) {};
    int readPacket();
    void sendJSONPacket(JsonObject& cmd);
    void sendJSONPacket(JsonObject& cmd, String strIP, uint16_t port);
    void sendJSONPacket(JsonObject& cmd, IPAddress ip, uint16_t port);
    void sendStringPacket(const char *cmd);
    void sendStringPacket(const char *cmd, String strIP, uint16_t port);
    void sendStringPacket(const char *cmd, IPAddress ip, uint16_t port);
    char packetBuffer[MAX_PACKET_SIZE];
  private:
    void cleanBuffer();
    WiFiUDP _udp;
    IPAddress _ip;
    uint16_t _port;
};


// Basic structure, all devices + gateways are linked to it
class XiaomiGenericDevice {
  public:
    XiaomiGenericDevice(String model, String sid, String short_id): model(model), sid(sid), short_id(short_id) {};
    void updateDevice(String model, String sid, String short_id);
    virtual void updateData(JsonObject& data) {}
    virtual bool prepareDataForSendCommand(XIAOMI_COMMAND deviceCommand, JsonObject& data) { return false; }

    String model;
    String sid;
    String short_id;
};

// Helper structure to hold SID and device instance
struct XiaomiDeviceNode {
  String sid;
  XiaomiGenericDevice *device = NULL;
};

class XiaomiBatteryDevice : public XiaomiGenericDevice {
  public:
    XiaomiBatteryDevice(String model, String sid, String short_id): XiaomiGenericDevice(model, sid, short_id) {};
    void updateData(JsonObject& data);
    
    float voltage;
};

class XiaomiCubeDevice : public XiaomiBatteryDevice {
  public:
    XiaomiCubeDevice(String model, String sid, String short_id): XiaomiBatteryDevice(model, sid, short_id) {};
    
    static constexpr const char *DEVICE_MODEL = "cube";
};

class XiaomiGatewayDevice : public XiaomiGenericDevice {
  public:
    XiaomiGatewayDevice(String model, String sid, String short_id): XiaomiGenericDevice(model, sid, short_id),
                        brightness(100), rgb((RGB){255, 255, 255}), illumination(0) {};
    void updateData(JsonObject& data);
    bool prepareDataForSendCommand(XIAOMI_COMMAND deviceCommand, JsonObject& data);

    uint8_t brightness;
    RGB rgb;
    uint16_t illumination;
    
    static constexpr const char *DEVICE_MODEL = "gateway";
};

class XiaomiMagnetDevice : public XiaomiBatteryDevice {
  public:
    XiaomiMagnetDevice(String model, String sid, String short_id): XiaomiBatteryDevice(model, sid, short_id) {};
    
    static constexpr const char *DEVICE_MODEL = "magnet";
};

class XiaomiMotionDevice : public XiaomiBatteryDevice {
  public:
    XiaomiMotionDevice(String model, String sid, String short_id): XiaomiBatteryDevice(model, sid, short_id) {};
    
    static constexpr const char *DEVICE_MODEL = "motion";
};

class XiaomiSensorHTDevice : public XiaomiBatteryDevice {
  public:
    XiaomiSensorHTDevice(String model, String sid, String short_id): XiaomiBatteryDevice(model, sid, short_id) {};
    void updateData(JsonObject& data);

    float temperature;
    float humidity;
    
    static constexpr const char *DEVICE_MODEL = "sensor_ht";
};

class XiaomiSensorWleakAQ1Device : public XiaomiBatteryDevice {
  public:
    XiaomiSensorWleakAQ1Device(String model, String sid, String short_id): XiaomiBatteryDevice(model, sid, short_id) {};

    static constexpr const char *DEVICE_MODEL = "sensor_wleak.aq1";
};

class XiaomiSmokeDevice : public XiaomiBatteryDevice {
  public:
    XiaomiSmokeDevice(String model, String sid, String short_id): XiaomiBatteryDevice(model, sid, short_id) {};
    
    static constexpr const char *DEVICE_MODEL = "smoke";
};

class XiaomiSwitchDevice : public XiaomiBatteryDevice {
  public:
    XiaomiSwitchDevice(String model, String sid, String short_id): XiaomiBatteryDevice(model, sid, short_id) {};
    
    static constexpr const char *DEVICE_MODEL = "switch";
};


// Xiaomi Gateway
class XiaomiGateway {
  public:
    XiaomiGateway(String ip, uint16_t port, String sid, String password, XiaomiUDP *udpMC, XiaomiCallback callBack):
                  ip(ip), port(port), sid(sid), password(password), _udpMC(udpMC), _callBack(callBack) {};
    void discoverDevices();
    void updateKey(String token);

    void addDevice(String sid);
    void updateDevice(String sid, String model, String short_id);
    bool deviceExists(String sid);
    XiaomiGenericDevice *getDeviceBySID(String sid);

    void readDevice(String sid);
    void sendCommandToDevice(String sid, XIAOMI_COMMAND deviceCommand);
    
    String ip;
    uint16_t port;
    String sid;

    unsigned int numDevices = 0;
    XiaomiDeviceNode devices[MAX_DEVICES];
  private:
    XiaomiUDP *_udpMC;
    XiaomiCallback _callBack;
    
    String password;
    char key[16*2 + 1]; // 33 bytes in total to store hex key
};


class XiaomiInterface {
  public:
    XiaomiInterface(XIAOMI_GATEWAYS definedGateways[], int numDefinedGateways, XiaomiCallback callBack);
    void discoverGateways();
    void discoverDevices();
    void readAllDevices();

    bool processPackets();
    void processAllPackets();

    void sendCommandToDevice(String model, String sid, XIAOMI_COMMAND deviceCommand);
    XiaomiGenericDevice *getDeviceBySID(String sid);
    XiaomiGateway *getGatewayForDeviceBySID(String sid);
    
  protected:
    XiaomiGateway *getGatewayByIPAddress(String ip);
    void addGateway(String ip, uint16_t port, String sid, String key);
  private:
    XiaomiUDP *_udpMC;
    XiaomiUDP *_udpServer;
    XIAOMI_GATEWAYS _definedGateways[MAX_GATEWAYS];
    XiaomiCallback _callBack;

    unsigned int numGateways = 0;
    XiaomiGateway *gateways[MAX_GATEWAYS];
};

#endif
